@@ -8,7 +8,7 @@ package XBase::SQL;
 use strict;
 use vars qw( $VERSION %COMMANDS );
 
-$VERSION = '0.155';
+$VERSION = '0.162';
 
 # #################################
 # Type conversions for create table
@@ -89,7 +89,7 @@ my %TYPES = ( 'char' => 'C', 'varchar' => 'C',
 	'WHERE' =>	'where WHEREEXPR',
 	'WHEREEXPR' =>	'BOOLEAN',
 
-	'BOOLEAN' =>	q'( \( BOOLEAN \) | RELATION ) ( ( AND | OR ) BOOLEAN ) *',
+	'BOOLEAN' =>	q'not BOOLEAN | ( \( BOOLEAN \) | RELATION ) ( ( AND | OR ) BOOLEAN ) *',
 	'RELATION' =>   'ARITHMETIC ( is not ? null | LIKE CONSTANT_NOT_NULL | RELOP ARITHMETIC )',
 	'AND' =>	'and',
 	'OR' =>		'or',
@@ -287,7 +287,7 @@ my %STORE = (
 	'WHEREEXPR' => sub { my $self = shift;
 		my $expr = join ' ', get_strings(@_);
 		### print STDERR "Evalling: $expr\n";
-		use Data::Dumper;
+		### use Data::Dumper;
 		my $fn = eval '
 			sub { 
 			### print Dumper @_;
@@ -389,7 +389,7 @@ sub parse
 		$self->store_results(\@result, \%STORE);
 		if (defined $self->{'whereerror'})
 			{ $self->{'errstr'} = "Some deeper problem: eval failed: $self->{'whereerror'}"; }
-		use Data::Dumper; print STDERR "Parsed $string to\n", Dumper $self if $ENV{'SQL_DUMPER'};
+		### use Data::Dumper; print STDERR "Parsed $string to\n", Dumper $self if $ENV{'SQL_DUMPER'};
 		}
 	$self;
 	}
@@ -659,6 +659,8 @@ use overload
 	'==' => sub { my $a = shift->notequal(@_); return ( $a ? 0 : 1); },
 	'""' => sub { ref shift; },
 	'.' => sub { XBase::SQL::Expr->string($_[0]->value . $_[1]->value); },
+	'*'  => sub { XBase::SQL::Expr->number($_[0]->value * $_[1]->value);},
+	'not'  => sub { not $_[0]->value },
 	;
 
 sub new { bless {}, shift; }
@@ -769,7 +771,7 @@ sub likematch
 
 	my $regexp = $string->value;
 	$regexp =~ s/(\\\\[%_]|.)/ ($1 eq '%') ? '.*' : ($1 eq '_') ? '.' : "\Q$1" /seg;
-	$field->value =~ /^$regexp$/i;
+	$field->value =~ /^$regexp$/si;
 	}
 
 1;
